from collections import defaultdict, Counter
from dataclasses import dataclass
from enum import Enum
from random import shuffle, randrange
from typing import List, Dict

from model.animals import random_animal_name


class Value(Enum):
    Two = 1
    Three = 2
    Four = 3
    Five = 4
    Six = 5
    Seven = 6
    Eight = 7
    Nine = 8
    Ten = 9
    Jack = 10
    Queen = 11
    King = 12
    Ace = 13


class Color(Enum):
    Hearts = "♥"
    Spades = "♠"
    Clubs = "♣"
    Diamonds = "♦"


@dataclass(frozen=True)
class Card:
    value: Value
    color: Color

    def __cmp__(self, other: "Card"):
        my = self.score()
        their = other.score()
        return (my > their) - (my < their)

    def __lt__(self, other: "Card"):
        return (self.score() < other.score())

    def __gt__(self, other: "Card"):
        return (self.score() > other.score())

    def __eq__(self, other: "Card"):
        return (self.score() == other.score())

    def __str__(self) -> str:
        return f"{self.value.value} {self.color.value}"

    def score(self) -> int:
        return int(self.value.value)


class Hand:
    def __init__(self, cards: List[Card] = None):
        if cards is None:
            cards = []
        self.cards: List[Card] = cards

    def __len__(self) -> int:
        return len(self.cards)

    def __getitem__(self, item):
        return self.cards[item]

    def __repr__(self):
        return "|".join([str(c) for c in self.cards])

    def give(self, other: Card):
        self.cards.append(other)

    def value(self):
        counter = Counter([c.value for c in self.cards])
        print("Counter:", counter)

        has_pair = False
        has_double_pair = False
        has_brelan = False
        has_full = False
        has_carre = False

        highest_card = None
        pair = None
        double_pair = None
        brelan = None
        full_des = None
        full_par = None
        carre = None

        for element, count in counter.items():
            element_cards = [c for c in self.cards if c.value == element]
            card = element_cards[0]  # Note we take a random color
            highest_card = max(highest_card, card) if highest_card else card

            if count == 2:
                if has_pair:
                    has_double_pair = True
                    double_pair = max(pair, card)
                    pair = min(pair, card)
                else:
                    has_pair = True
                    pair = max(pair, card) if pair else card
            if count == 3:
                has_brelan = True
                brelan = max(brelan, card) if brelan else card
            if has_brelan and has_pair:
                has_full = True
                full_des = max(brelan, card) if brelan else card
                full_par = max(pair, card) if pair else card
            if count == 4:
                has_carre = True
                carre = max(carre, card) if carre else card

        print(" | ".join([
            f"ANALYSIS",
            f"Carre[{carre}]" if has_carre else "no carre",
            f"Full[{full_des}|{full_par}]" if has_full else "no full",
            f"Brelan[{brelan}]" if has_brelan else "no carre",
            f"Double paire[{double_pair}|{pair}]" if has_double_pair else "no Dpaire",
            f"Paire[{pair}]" if has_pair else "no paire",
            f"Card[{highest_card}]"
        ]), end=" ")
        # Finished counting, let's return scores
        if has_carre:
            score = (20 ** 5) * carre.score()
        elif has_full:
            score = (20 ** 4) * full_des.score() + full_par.score()
        elif has_brelan:
            score = (20 ** 3) * brelan.score()
        elif has_double_pair:
            score = (20 ** 2) * double_pair.score() + pair.score()
        elif has_pair:
            score = 20 * pair.score()
        else:
            score = highest_card.score()

        print("\t-> score=\t", score)
        return score


class Deck:
    def __init__(self):
        self.cards = [Card(v, c) for v in Value for c in Color]
        self.defausse = []

    def __len__(self):
        return len(self.cards)

    def random_card(self) -> Card:
        if not self.cards:
            self.reset()

        card = self.cards.pop(randrange(len(self.cards)))
        self.defausse.append(card)
        return card

    def reset(self):
        self.cards.extend(self.defausse)
        self.defausse.clear()
        shuffle(self.cards)


class Player:
    name: str = random_animal_name()
    hand: Hand = Hand()

    def give(self, card: Card):
        self.hand.give(card)


class Game:
    def __init__(self, deck: Deck = Deck(), players=None):
        if players is None:
            players = []

        self.deck: Deck = deck
        self.players: List[Player] = players
        self.defeats: Dict[Player, int] = defaultdict(lambda: 0)

    def new_game(self):
        self.deck.reset()
        while len(self.players) > 1:
            self.new_turn()
        winner = self.players[0]
        print(f"Game over - Player {winner.name} wins with {len(winner.hand)} cards!")
        pass

    def new_turn(self):
        # Distribution
        shuffle(self.deck)
        for p in self.players:
            p.give(self.deck.random_card())

        # Tour
        for p in self.players:
            announce = p.announce(current_bet)

        # TODO: Put next first player first of list
