from unittest import TestCase

from model.data import Deck, Player, Card, Value, Color, Hand
from model.hands import brelan, pair, single, double_pair

ACE_OF_HEARTS = Card(Value.Ace, Color.Hearts)
ACE_OF_SPADES = Card(Value.Ace, Color.Spades)
ACE_OF_DIAMONDS = Card(Value.Ace, Color.Diamonds)
ACE_OF_CLUBS = Card(Value.Ace, Color.Clubs)

PAIR_ACE = pair(Value.Ace)
SINGLE_ACE = single(Value.Ace)
DOUBLE_PAIR_ACE = double_pair(Value.Ace)


class TestDeck(TestCase):

    def setUp(self) -> None:
        super().setUp()
        self.deck = Deck()

    def testInitDeck(self):
        self.assertEqual(52, len(self.deck), "52 cards")
        self.assertEqual(52, len(set(self.deck.cards)), "unique cards")

    def testRandomCard(self):
        cards = []
        for i in range(len(self.deck)):
            cards.append(self.deck.random_card())

        self.assertEqual(52, len(set(cards)), "draw each card once")


class TestPlayer(TestCase):

    def setUp(self) -> None:
        super().setUp()
        self.player = Player()

    def testHand(self):
        self.assertEqual(0, len(self.player.hand), "Begin no cards")
        self.assertTrue(len(self.player.name), "Has a name")

        self.player.give(ACE_OF_HEARTS)

        self.assertEqual(1, len(self.player.hand), "Gave one card")
        self.assertEqual(Value.Ace, self.player.hand[0].value, "Is Ace")
        self.assertEqual(Color.Hearts, self.player.hand[0].color, "of Hearts")

    def testDefeats(self):
        pass


def lowest_card_and_rest():
    lowValue: Value = Value.Two
    otherValues = list(Value)
    otherValues.remove(lowValue)

    return lowValue, otherValues


class TestHand(TestCase):
    def setUp(self) -> None:
        self.hand = Hand()

    def testSimple(self):
        low_value, other_values = lowest_card_and_rest()

        for value in other_values:
            high_hand = single(value)
            low_hand = single(low_value)

            self.assertGreater(high_hand.value(), low_hand.value())

    def testPair(self):
        low_value, other_values = lowest_card_and_rest()

        for value in other_values:
            high_hand: Hand = pair(value)
            low_hand: Hand = pair(low_value)
            single_hand: Hand = single(Value.Ace)

            high_score = high_hand.value()
            low_score = low_hand.value()
            single_score = single_hand.value()

            self.assertGreater(high_score, low_score, f"Pair[{high_hand}] > Pair[{low_hand}]]")
            self.assertGreater(high_score, single_score, f"Pair[{high_hand}] > Ace")

    def testDoublePair(self):
        low_value, other_values = lowest_card_and_rest()

        for value in other_values:
            high_hand: Hand = double_pair(value)
            low_hand: Hand = double_pair(low_value)
            pair_hand: Hand = PAIR_ACE
            single_hand: Hand = SINGLE_ACE

            high_score = high_hand.value()
            low_score = low_hand.value()
            pair_score = pair_hand.value()
            single_score = single_hand.value()

            # Consider case when we compare equal double pairs
            low_cards = [c.value for c in low_hand.cards]
            if all([card.value in low_cards for card in high_hand.cards]):
                self.assertEqual(high_score, low_score, f"DoublePair[{high_hand}] == DoublePair[{low_hand}]")
            else:
                self.assertGreater(high_score, low_score, f"DoublePair[{high_hand}] > DoublePair[{low_hand}]")
            self.assertGreater(high_score, pair_score, f"DoublePair[{high_hand}] > Pair[Ace]")
            self.assertGreater(high_score, single_score, f"DoublePair[{high_hand}] > Ace")

    def testBrelan(self):
        low_value, other_values = lowest_card_and_rest()

        for value in other_values:
            high_hand: Hand = brelan(value)
            low_hand: Hand = brelan(low_value)
            double_pair_hand: Hand = DOUBLE_PAIR_ACE
            pair_hand: Hand = PAIR_ACE
            single_hand: Hand = SINGLE_ACE

            high_score = high_hand.value()
            low_score = low_hand.value()
            double_pair_score = double_pair_hand.value()
            pair_score = pair_hand.value()
            single_score = single_hand.value()

            self.assertGreater(high_score, low_score, f"Brelan[{high_hand}] > Brelan[{low_hand}]]")
            self.assertGreater(high_score, double_pair_score, f"Brelan[{high_hand}] > Pair[Ace]")
            self.assertGreater(high_score, pair_score, f"Brelan[{high_hand}] > Pair[Ace]")
            self.assertGreater(high_score, single_score, f"Brelan[{high_hand}] > Ace")
